/*******************************************************************************
 *   Copyright (C) 2007, 2008, 2009, 2010 Peter Kolb
 *   peter.kolb@linguatools.org
 *
 *   Licensed under the Apache License, Version 2.0 (the "License"); you may not
 *   use this file except in compliance with the License. You may obtain a copy
 *   of the License at 
 *   
 *        http://www.apache.org/licenses/LICENSE-2.0 
 *
 *   Unless required by applicable law or agreed to in writing, software 
 *   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the 
 *   License for the specific language governing permissions and limitations
 *   under the License.
 *
 ******************************************************************************/

package de.linguatools.disco;

import java.io.*;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.Directory;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.Query;
import org.apache.lucene.analysis.WhitespaceAnalyzer;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Hits;
import org.apache.lucene.document.Document;

/*******************************************************************************
 * DISCO (Extracting DIStributionally Similar Words Using CO-occurrences) provides
 * a number of methods for computing the distributional (i.e. semantic) similarity
 * between arbitrary words, for retrieving a word's collocations or its corpus
 * frequency. It also provides a method to retrieve the semantically most similar
 * words for a given word.
 * @author peter
 *******************************************************************************/
public class DISCO {

    /***************************************************************************
     * Searches for a word in index field "word" and returns the first hit
     * Document or null.<br/>
     * DISCO uses the <a href="http://lucene.apache.org">Lucene</a> index. A
     * word's data are stored in the index in an object of type Document. A 
     * Document has the following 16 fields:<ul>
     * <li>"word": contains a word, tokenized with WhitespaceAnalyzer. This is
     * the only searchable field.</li>
     * <li>"freq": the corpus frequency of the word. This field is only stored,
     * but not indexed.</li>
     * <li>"dsb": the distributionally similar words for word. They are stored
     * in a single string, in which the words are separated by spaces. This
     * field is not indexed, and therefore not searchable.</li>
     * <li>"dsbSim": contains a single string with the similarity values for the
     * words in the field "dsb", separated by spaces. The string in this field 
     * is parallel to the string in the field "dsb", i.e., the n-th value of the
     * string in "dsbSim" corresponds to the n-th word in "dsb".<br/>
     * <b>To get correct values precede each value string with "0." before 
     * converting it to float!</b> (the leading "0." of all values have been
     * deleted in order to reduce index size).<br/>
     * Example: field "dsb" contains the string "apple banana cherry", field 
     * "dsbSim" contains the string "3241 1233 0788". This means that the
     * similarity between the word in the field "word" and cherry is 0.0788.
     * </li>
     * <li>"kol1": contains all words that have occurred three words left to
     * "word" in the corpus (collocations). Stored as space-separated single
     * string.</li>
     * <li>"kol1Sig": contains the significance values for "kol1", in a string
     * parallel to the string in "kol1".</li>
     * <li>"kol2": same as "kol1" but two words to the left.</li>
     * <li>"kol2Sig": same as "kol1Sig".</li>
     * <li>...</li>
     * <li>"kol6": same as "kol1" but three words to the right.</li>
     * <li>"kol6Sig" </li>
     * </ul>
     * @param indexName name of index directory
     * @param word word to be looked up in index
     * @return index entry of word or null
     */
    public Document searchIndex(String indexName, String word)
            throws IOException{
        try {
            // hole den Inhalt von indexDir
            Directory fsDir = FSDirectory.getDirectory(indexName);
            // erzeuge einen IndexSearcher fuer das indexDir
            IndexSearcher is = new IndexSearcher(fsDir);
            // Tokenizer definieren
            WhitespaceAnalyzer analyzer = new WhitespaceAnalyzer();
            // Anfrage tokenisieren und parsen
            Query query;
            QueryParser parser = new QueryParser("word", analyzer);
            query = parser.parse(word); // can throw ParseException !
            // nach Anfrage im Index suchen
            Hits hits = is.search(query);
            if( hits.length() == 0 ) return null;
            // Nur den ersten Treffer verwenden (es sollte nur einen geben)
            Document doc = hits.doc(0);
            return doc;
        } catch (ParseException ex) {
            System.err.println("Error: ParseException: "+ex);
            return null;
        }
    }

    /***************************************************************************
     * returns the number of Documents (i.e. words) in the index.
     * @param indexName name of index directory
     * @return number of words in index
     * @throws java.io.IOException
     */
    public int numberOfWords(String indexName) throws IOException{
        // hole den Inhalt von indexDir
        Directory fsDir = FSDirectory.getDirectory(indexName);
        // erzeuge einen IndexReader fuer das indexDir
        IndexReader ir = IndexReader.open( fsDir );
        // Hole Anzahl Dokumente im Index
        return ( ir.numDocs() );
    }

    /***************************************************************************
     * Looks up the input word in the index and returns its frequency.
     * If the word is not found the return value is zero.
     * @param indexName name of index directory
     * @param word word to be looked up
     * @return frequency of the input word
     * @throws java.io.IOException
     */
    public int frequency(String indexName, String word) throws IOException{
        Document doc = searchIndex(indexName, word);
        if ( doc == null ) return 0;
        return Integer.parseInt( doc.get("freq") );
    }

    /***************************************************************************
     * Looks up the input word in the index and returns its distributionally
     * similar words ordered by decreasing similarity together with similarity
     * values. If the search word isn't found in the index, the return value is
     * null.<br/>
     * <a href="http://www.cs.ualberta.ca/~lindek/papers/acl98.pdf">Lin</a>'s
     * similarity measure was used to compute the similar words.
     * @param indexName name of index directory
     * @param word word to be looked up
     * @return result data structure or null
     * @throws IOException
     */
    public ReturnDataBN similarWords(String indexName, String word)
            throws IOException{
        Document doc = searchIndex(indexName, word);
        if ( doc == null ) return null;
        // die komprimiert gespeicherten Inhalte der Felder "dsb" und "dsbSim"
        // holen
        ReturnDataBN res = new ReturnDataBN();
        res.words = doc.get("dsb").split(" ");
        res.values = doc.get("dsbSim").split(" ");
        return res;
    }

    /***************************************************************************
     * Returns the collocations for the input word together with their
     * significance values, ordered by significance value (highest significance
     * first). If the search word isn't found in the index, the return value is
     * null.<br/>
     * The significance values were computed using <a
     * href="http://www.cs.ualberta.ca/~lindek/papers/acl98.pdf">Lin</a>'s
     * measure.
     * @param indexName name of the index directory
     * @param word the input word
     * @return the list of collocations with their significance values
     * @throws java.io.IOException
     */
    public ReturnDataCol[] collocations(String indexName, String word)
            throws IOException{

        Document doc = searchIndex(indexName, word);
        if ( doc == null ) return null;
        // die komprimiert gespeicherten Inhalte der Felder "Kol1" und "KolSig1"
        // bis "Kol6" und "KolSig6" holen und in ein Hash speichern. Die
        // einzelnen Relationen (1-6) werden zusammengefasst und die jeweiligen
        // Signifikanzwerte addiert.
        Hashtable colloHash = new Hashtable();
        String[] wordsBuffer;
        String[] valuesBuffer;
        int i; int rel;
        float sig;
        for ( rel = 1; rel <= 6; rel++ ){
            wordsBuffer = doc.get("kol"+Integer.toString(rel)).split(" ");
            valuesBuffer = doc.get("kol"+Integer.toString(rel)+"Sig").split(" ");
            for( i = 1; i < wordsBuffer.length; i++ ){
                if( colloHash.get(wordsBuffer[i]) == null ){
                    colloHash.put(wordsBuffer[i],
                            Float.parseFloat(valuesBuffer[i]) );
                }else{
                    sig = (float) (Float.parseFloat(valuesBuffer[i])) +
                            ((Float) colloHash.get(wordsBuffer[i]));
                    colloHash.put(wordsBuffer[i], sig);
                }
            }
        }
        // jetzt das Hash in ein Array speichern und nach Signifikanz sortieren
        ReturnDataCol[] res = new ReturnDataCol[colloHash.size()];
        Enumeration e = colloHash.keys();
        i = 0;
        while( e.hasMoreElements() ){
            String w = (String) e.nextElement();
            res[i++] = new ReturnDataCol(w,
                    ((Float)colloHash.get(w)).floatValue() );
        }
        // sortiere Array ReturnDataCol[] nach hoechstem Signifikanzwert
        Arrays.sort( res, new ValueComparator() );
        return res;
    }

    /***************************************************************************
     * Computes the first order similarity (according to <a
     * href="http://www.cs.ualberta.ca/~lindek/papers/acl98.pdf">Lin</a>'s
     * vector similarity measure) between the input words based on their
     * collocation sets. If any of the two words isn't found in the index, the
     * return value is -1.
     * @param indexName name of index directory
     * @param w1 input word #1
     * @param w2 input word #2
     * @return similarity value
     * @throws java.io.IOException
     */
    public float firstOrderSimilarity(String indexName, String w1, String w2)
            throws IOException{
        // die beiden zu vergleichenden Wörter im Index nachschlagen
        Document doc1 = searchIndex(indexName, w1);
        Document doc2 = searchIndex(indexName, w2);
        if ( doc1 == null || doc2 == null ) return -1;
        // Kollokationen von Wort #1 durchlaufen (über alle Relationen), in Hash
        // speichern (nach Relationen unterschieden) und alle Werte addieren.
        Hashtable colloHash = new Hashtable();
        String[] wordsBuffer;
        String[] valuesBuffer;
        int i; int rel;
        float nenner = 0, v;
        for ( rel = 1; rel <= 6; rel++ ){
            wordsBuffer = doc1.get("kol"+Integer.toString(rel)).split(" ");
            valuesBuffer = doc1.get("kol"+Integer.toString(rel)+"Sig").split(" ");
            for( i = 1; i < wordsBuffer.length; i++ ){
                v = Float.parseFloat(valuesBuffer[i]);
                colloHash.put(wordsBuffer[i]+"_"+Integer.toString(rel), v );
                nenner += v;
            }
        }
        // Kollokationen von Wort #2 durchlaufen (über alle Relationen), mit den
        // Kollokationen von Wort #1 im Hash vergleichen und ggf. die Werte zum
        // Zähler addieren und alle Werte zum Nenner addieren.
        float zaehler = 0;
        for ( rel = 1; rel <= 6; rel++ ){
            wordsBuffer = doc2.get("kol"+Integer.toString(rel)).split(" ");
            valuesBuffer = doc2.get("kol"+Integer.toString(rel)+"Sig").split(" ");
            for( i = 1; i < wordsBuffer.length; i++ ){
                v = Float.parseFloat(valuesBuffer[i]);
                if ( colloHash.containsKey(wordsBuffer[i]+"_"+Integer.toString(rel)) ){
                    zaehler += v + (Float)
                            colloHash.get(wordsBuffer[i]+"_"+Integer.toString(rel));
                }
                nenner += v;
            }
        }
        return zaehler / nenner;
    }

    /***************************************************************************
     * Computes the second order similarity (according to Lin's measure) between
     * the input words based on the sets of their distributional similar words.
     * If any of the two words isn't found in the index, the return value is -1.
     * @param indexName name of index directory
     * @param w1 input word #1
     * @param w2 input word #2
     * @return similarity value
     * @throws java.io.IOException
     */
    public float secondOrderSimilarity(String indexName, String w1, String w2)
            throws IOException{
        // die beiden zu vergleichenden Wörter im Index nachschlagen
        Document doc1 = searchIndex(indexName, w1);
        Document doc2 = searchIndex(indexName, w2);
        if ( doc1 == null || doc2 == null ) return -1;
        // ähnliche Wörter von Wort #1 durchlaufen, in Hash speichern und alle
        // Werte addieren.
        Hashtable simHash = new Hashtable();
        String[] wordsBuffer;
        String[] valuesBuffer;
        int i;
        float nenner = 0, v;
        wordsBuffer = doc1.get("dsb").split(" ");
        valuesBuffer = doc1.get("dsbSim").split(" ");
        for( i = 1; i < wordsBuffer.length; i++ ){
            // aus Speicherplatzgründen wurde beim Indexieren die führende "0."
            // weggelassen -- wieder anfügen:
            v = Float.parseFloat("0."+valuesBuffer[i]);
            simHash.put(wordsBuffer[i], v );
            nenner += v;
        }
        // ähnliche Wörter von Wort #2 durchlaufen, mit den ähnlichen Wörtern von
        // Wort #1 im Hash vergleichen und ggf. die Werte zum Zähler addieren und
        // alle Werte zum Nenner addieren.
        float zaehler = 0;
        wordsBuffer = doc2.get("dsb").split(" ");
        valuesBuffer = doc2.get("dsbSim").split(" ");
        for( i = 1; i < wordsBuffer.length; i++ ){
            // beim Indexieren weggelassene "0." wieder anfügen
            v = Float.parseFloat("0."+valuesBuffer[i]);
            if ( simHash.containsKey(wordsBuffer[i]) ){
                zaehler += v + (Float) simHash.get(wordsBuffer[i]);
            }
            nenner += v;
        }
        return zaehler / nenner;
    }

}
